/* 
 * Copyright 1990 Richard Hesketh / rlh2@ukc.ac.uk
 *                Computing Lab. University of Kent at Canterbury, UK
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of Richard Hesketh and The University of
 * Kent at Canterbury not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Richard Hesketh and The University of Kent at Canterbury make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * Richard Hesketh AND THE UNIVERSITY OF KENT AT CANTERBURY DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Richard Hesketh OR THE
 * UNIVERSITY OF KENT AT CANTERBURY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 *
 * Author:  Richard Hesketh / rlh2@ukc.ac.uk, 
 *                Computing Lab. University of Kent at Canterbury, UK
 */

#include<ctype.h>
#include <stdio.h>
#include <stdlib.h>

#include<X11/Xlib.h>
#include<X11/Xatom.h>
#include<X11/Xos.h>
#include<X11/IntrinsicP.h>
#include<X11/StringDefs.h>
#include<X11/Xmu/CurUtil.h>

#include "Convert.h"

extern void XmuCopyISOLatin1Lowered();
extern Pixmap XukcLocatePixmapFile();

#define ErrorMessage(routine, message) { \
	Cardinal np = 0; \
	XtAppErrorMsg(XtDisplayToApplicationContext(display), \
			"wrongParameters", routine, "XukcToolkitError", \
			message, (String *)NULL, &np); \
}

/* copied from Xt/Converters.c !
 */
#define done(type, value) \
	{                                                       \
	    if (toVal->addr != NULL) {                          \
		if (toVal->size < sizeof(type)) {		\
		    toVal->size = sizeof(type);			\
		    return False;                               \
		}						\
		*(type*)(toVal->addr) = (value);		\
	    } else {						\
		static type static_val;				\
		static_val = (value);				\
		toVal->size = sizeof(type);			\
		toVal->addr = (XtPointer)&static_val;		\
	    }							\
	    return True;					\
	}

#define done_string(value) \
	{                                                       \
	    Cardinal size;					\
	    if (value == NULL) /* we shouldn't have caught this */	\
		return False;					\
	    size = strlen(value) + 1;				\
	    if (toVal->addr != NULL) {                          \
		if (toVal->size < size) {			\
		    toVal->size = size;				\
		    return False;                               \
		}						\
		(void)strcpy((char *)toVal->addr, value);	\
		toVal->size = size;				\
	    } else {						\
		static String static_val;			\
		static_val = (value);				\
		toVal->size = size;				\
		toVal->addr = (XtPointer)&static_val;		\
	    }							\
	    return True;					\
	}

XtConvertArgRec screen_Args[] = {
  {XtWidgetBaseOffset, (XtPointer)XtOffset(Widget, core.screen),
   sizeof(Screen*)}
};

XtConvertArgRec pixmap_Args[] = {
  {XtWidgetBaseOffset, (XtPointer)XtOffset(Widget, core.screen),
   sizeof(Screen*)},
  {XtWidgetBaseOffset, (XtPointer)XtOffset(Widget, core.widget_class),
   sizeof(WidgetClass)},
#ifdef UKC_WIDGETS
  {XtWidgetBaseOffset, (XtPointer)XtOffset(UKCLabelWidget, label.foreground),
   sizeof(Pixel)},
  {XtWidgetBaseOffset, (XtPointer)XtOffset(Widget, core.background_pixel),
   sizeof(Pixel)}
#endif /* UKC_WIDGETS */
};

Boolean
XukcCvtStringToPixmap(display, args, num_args, fromVal, toVal, conv_data)
Display *display;
XrmValuePtr args;
Cardinal    *num_args;
XrmValuePtr	fromVal;
XrmValuePtr	toVal;
XtPointer 	*conv_data;
{
    static Pixmap pixmap;		/* static for cvt magic */
    char *name = (char *)fromVal->addr;
    Pixel fg, bg;
    Screen *screen;

    if (!(*num_args == 2 || *num_args == 4))
	ErrorMessage("cvtStringToPixmap",
	     "String to Pixmap conversion needs screen argument and optional foreground and background colours");

    if (conv_data != NULL)
	*conv_data = NULL;

    if (strcmp(name, "Unspecified") == 0) {
	pixmap = XtUnspecifiedPixmap;
	done(Pixmap, pixmap);
    }

    if (strcmp(name, "None") == 0) {
	pixmap = None;
	done(Pixmap, pixmap);
    }

    if (strcmp(name, "ParentRelative") == 0) {
	pixmap = ParentRelative;
	done(Pixmap, pixmap);
    }

    screen = *((Screen **)args[0].addr);
    screen || (screen = DefaultScreenOfDisplay(display));

#ifdef UKC_WIDGETS
    if (*num_args == 4 &&
	*((WidgetClass *)args[1].addr) == ukcLabelWidgetClass) {
		fg = *((Pixel *)args[2].addr);
		bg = *((Pixel *)args[3].addr);
    } else {
		fg = BlackPixelOfScreen(screen);
		bg = WhitePixelOfScreen(screen);
    }
#else
    fg = BlackPixelOfScreen(screen);
    bg = WhitePixelOfScreen(screen);
#endif /* UKC_WIDGETS */

    pixmap = XukcLocatePixmapFile(screen, name, NULL, 0, NULL, NULL, fg, bg);

    if (pixmap != None) {
#if 0
	if (conv_data != NULL)
		*conv_data = save_pixmap_cvt(name, pixmap);
#endif
	done(Pixmap, pixmap);
    }

    XtDisplayStringConversionWarning(display, name, "Pixmap");
    return FALSE;
}


/* The following has been taken from Xmu/StrToBmap.c and Xmu/StrCurs.c
 * and then embellished a bit .. to support caching of names so that
 * they can be found and used by the added PixmapToString and CursorToString
 * converters.
 */

/***********************************************************
Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#include	<X11/Xmu/Drawing.h>

/*
 * XmuConvertStringToBitmap:
 *
 * creates a depth-1 Pixmap suitable for window manager icons.
 * "string" represents a bitmap(1) filename which may be absolute,
 * or relative to the global resource bitmapFilePath, class
 * BitmapFilePath.  If the resource is not defined, the default
 * value is the build symbol BITMAPDIR.
 *
 * shares lots of code with XmuConvertStringToCursor.  
 *
 * To use, include the following in your ClassInitialize procedure:

static XtConvertArgRec screenConvertArg[] = {
    {XtWidgetBaseOffset, (caddr_t) XtOffset(Widget, core.screen),
							sizeof(Screen *)}
};

    XtAddConverter("String", "Bitmap", XmuCvtStringToBitmap,
		   screenConvertArg, XtNumber(screenConvertArg));
 *
 */


static Boolean
XukcCvtStringToBitmap(display, args, num_args, fromVal, toVal, conv_data)
Display *display;
XrmValuePtr args;
Cardinal    *num_args;
XrmValuePtr	fromVal;
XrmValuePtr	toVal;
XtPointer 	*conv_data;
{
    static Pixmap pixmap;		/* static for cvt magic */
    char *name = (char *)fromVal->addr;

    if (*num_args != 1)
	ErrorMessage("cvtStringToBitmap",
		     "String to Bitmap conversion needs screen argument");

    *conv_data = NULL;
    /* these three are special cases */
    if (strcmp(name, "Unspecified") == 0) {
	pixmap = XtUnspecifiedPixmap;
	done(Pixmap, pixmap);
    }

    if (strcmp(name, "None") == 0) {
	pixmap = None;
	done(Pixmap, pixmap);
    }

    if (strcmp(name, "ParentRelative") == 0) {
	pixmap = ParentRelative;
	done(Pixmap, pixmap);
    }

    pixmap = XmuLocateBitmapFile (*((Screen **) args[0].addr), name,
				  NULL, 0, NULL, NULL, NULL, NULL);

    if (pixmap != None) {
#if 0
	*conv_data = save_pixmap_cvt(name, pixmap);
#endif
	done(Pixmap, pixmap);
    }

    XtDisplayStringConversionWarning(display, name, "Bitmap");
    return FALSE;
}


#include	<X11/CoreP.h>		/* just to do XtConvert() */
#include	<sys/param.h>		/* just to get MAXPATHLEN */
#ifndef MAXPATHLEN
#define MAXPATHLEN 256
#endif

/*
 * XmuConvertStringToCursor:
 *
 * allows String to specify a standard cursor name (from cursorfont.h), a
 * font name and glyph index of the form "FONT fontname index [[font] index]", 
 * or a bitmap file name (absolute, or relative to the global resource
 * bitmapFilePath, class BitmapFilePath).  If the resource is not
 * defined, the default value is the build symbol BITMAPDIR.
 *
 * shares lots of code with XmuCvtStringToPixmap, but unfortunately
 * can't use it as the hotspot info is lost.
 *
 * To use, include the following in your ClassInitialize procedure:

static XtConvertArgRec screenConvertArg[] = {
    {XtWidgetBaseOffset, (caddr_t) XtOffset(Widget, core.screen),
							sizeof(Screen *)}
};

    XtAddConverter("String", "Cursor", XmuCvtStringToCursor,      
		   screenConvertArg, XtNumber(screenConvertArg));
 *
 */


#ifndef BITMAPDIR
#define BITMAPDIR "/usr/include/X11/bitmaps"
#endif

#define FONTSPECIFIER		"FONT "

/*ARGSUSED*/
static Boolean
XukcCvtStringToCursor(display, args, num_args, fromVal, toVal, conv_data)
Display *display;
XrmValuePtr args;
Cardinal    *num_args;
XrmValuePtr	fromVal;
XrmValuePtr	toVal;
XtPointer	*conv_data;
{
    static Cursor cursor;		/* static for cvt magic */
    char *name = (char *)fromVal->addr;
    Screen *screen;
    register int i;
    char maskname[MAXPATHLEN];
    Pixmap source, mask;
    static XColor bgColor = {0, ~0, ~0, ~0};  /* XXX - make a resource */
    static XColor fgColor = {0, 0, 0, 0};     /* XXX - ditto */
    int xhot, yhot;
    int len;

    if (*num_args != 1)
	ErrorMessage("cvtStringToCursor",
		     "String to cursor conversion needs screen argument");

    screen = *((Screen **) args[0].addr);
    screen || (screen = DefaultScreenOfDisplay(display));

    if (strcmp(name, "None") == 0) {
	cursor = None;
#if 0
	*conv_data = save_cursor_cvt(name, cursor);
#endif
	done(Cursor, cursor);
    }

    if (0 == strncmp(FONTSPECIFIER, name, strlen(FONTSPECIFIER))) {
	char source_name[MAXPATHLEN], mask_name[MAXPATHLEN];
	int source_char, mask_char, fields;
	WidgetRec widgetRec;
	Font source_font, mask_font;
	XrmValue fromString, toFont;

	fields = sscanf(name, "FONT %s %d %s %d",
			source_name, &source_char,
			mask_name, &mask_char);
	if (fields < 2) {
	    XtDisplayStringConversionWarning(display, name, "Cursor");
	    return FALSE;
	}

	/* widgetRec is stupid; we should just use XtDirectConvert,
	 * but the names in Xt/Converters aren't public. */
	widgetRec.core.screen = screen;
	fromString.addr = source_name;
	fromString.size = strlen(source_name);
	XtConvert(&widgetRec, XtRString, &fromString, XtRFont, &toFont);
	if (toFont.addr == NULL) {
	    XtDisplayStringConversionWarning(display, name, "Cursor");
	    return FALSE;
	}
	source_font = *(Font*)toFont.addr;

	switch (fields) {
	  case 2:		/* defaulted mask font & char */
	    mask_font = source_font;
	    mask_char = source_char;
	    break;

	  case 3:		/* defaulted mask font */
	    mask_font = source_font;
	    mask_char = atoi(mask_name);
	    break;

	  case 4:		/* specified mask font & char */
	    fromString.addr = mask_name;
	    fromString.size = strlen(mask_name);
	    XtConvert(&widgetRec, XtRString, &fromString, XtRFont, &toFont);
	    if (toFont.addr == NULL) {
		XtDisplayStringConversionWarning( display, name, "Cursor" );
		return FALSE;
	    }
	    mask_font = *(Font*)toFont.addr;
	}

	cursor = XCreateGlyphCursor( DisplayOfScreen(screen), source_font,
				     mask_font, source_char, mask_char,
				     &fgColor, &bgColor );
#if 0
	*conv_data = save_cursor_cvt(name, cursor);
#endif
	done(Cursor, cursor);
    }

    i = XmuCursorNameToIndex (name);
    if (i != -1) {
	cursor = XCreateFontCursor (DisplayOfScreen(screen), i);
#if 0
	*conv_data = save_cursor_cvt(name, cursor);
#endif
	done(Cursor, cursor);
    }

    if ((source = XmuLocateBitmapFile (screen, name, 
				       maskname, (sizeof maskname) - 4,
				       NULL, NULL, &xhot, &yhot)) == None) {
	XtDisplayStringConversionWarning (display, name, "Cursor");
	return FALSE;
    }
    len = strlen (maskname);
    for (i = 0; i < 2; i++) {
	strcpy (maskname + len, i == 0 ? "Mask" : "msk");
	if ((mask = XmuLocateBitmapFile (screen, maskname, NULL, 0, 
					 NULL, NULL, NULL, NULL)) != None)
	  break;
    }

    cursor = XCreatePixmapCursor( DisplayOfScreen(screen), source, mask,
				  &fgColor, &bgColor, xhot, yhot );
    XFreePixmap( DisplayOfScreen(screen), source );
    if (mask != None) XFreePixmap( DisplayOfScreen(screen), mask );

#if 0
    *conv_data = save_cursor_cvt(name, cursor);
#endif
    done(Cursor, cursor);
}

void
XukcRegisterConverters()
{
	static Boolean initialized = FALSE;

	if (initialized)
		return;
	else
		initialized = TRUE;

	XtSetTypeConverter(XtRString, XtRPixmap, XukcCvtStringToPixmap,
			   (XtConvertArgRec *)pixmap_Args, 
			   XtNumber(pixmap_Args),
			   XtCacheByDisplay, (XtDestructor)0);
	XtSetTypeConverter(XtRString, XtRCursor, XukcCvtStringToCursor,
			   (XtConvertArgRec *)screen_Args, 
			   XtNumber(screen_Args),
			   XtCacheByDisplay, (XtDestructor)0);
	XtSetTypeConverter(XtRString, XtRBitmap, XukcCvtStringToBitmap,
			   (XtConvertArgRec*)screen_Args, 
			   XtNumber(screen_Args),
			   XtCacheByDisplay, (XtDestructor)0);
}
