/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: FileDialog.c,v 0.9.1.4 1991/12/16 20:50:37 leei Exp $";

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/dir.h>

#include <unistd.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/ShellP.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/List.h>

#include "proto.h"
#include "strstr.h"

#include "EzMenu.h"
#include "Popups.h"
#include "ScrollList.h"
#include "InstBelow.h"

#include "FileDialogP.h"

#ifndef S_ISDIR
#define S_ISDIR(m)      (((m) & S_IFMT) == S_IFDIR)
#endif	/* S_ISDIR(m) */

#define DLOG_NOTDONE   (0)
#define DLOG_SELECTED  (1)
#define DLOG_CANCELLED (2)

#define LONG_OFF(dlog)	((dlog)->long_listing? 6: 0)

static void actSelect __(( Widget, XEvent *, String *, Cardinal *));
static void actCancel __(( Widget, XEvent *, String *, Cardinal *));

static void actNewDirectory __(( Widget, XEvent *, String *, Cardinal *));
static void actUpdateFileList __(( Widget, XEvent *, String *, Cardinal *));
static void actToggleLong __(( Widget, XEvent *, String *, Cardinal *));
static void actToggleFilter __(( Widget, XEvent *, String *, Cardinal *));
static void actJumpFile __(( Widget, XEvent *, String *, Cardinal *));
static void actNextFile __(( Widget, XEvent *, String *, Cardinal *));
static void actDeleteFile __(( Widget, XEvent *, String *, Cardinal *));
static void actSearch __(( Widget, XEvent *, String *, Cardinal *));

static XtActionsRec actions[] = {
    /* {name, procedure} */
    {"fdSelect", actSelect},
    {"fdCancel", actCancel},
    /* The following are menu selections. */
    {"fdNewDirectory",		actNewDirectory},
    {"fdUpdateFileList",	actUpdateFileList},
    {"fdToggleFilter",		actToggleFilter},
    {"fdToggleLongListing",	actToggleLong},
    {"fdJumpFile",		actJumpFile},
    {"fdNextFile",		actNextFile},
    {"fdDeleteFile",		actDeleteFile},
    {"fdSearch",		actSearch}
};

static char translations[] = 
"None <Key> Return:	fdSelect()\n\
 Ctrl <Key> c:		fdCancel()\n\
";

#define offset(field) XtOffset(FileDialogWidget,field)
static XtResource resources[] = {
    { XtNlongListing, XtCLongListing, XtRBoolean, sizeof(Boolean),
	offset(dialog.long_listing), XtRImmediate, (XtPointer) False },
    { XtNsearchPath, XtCSearchPath, XtRString, sizeof(String),
	offset(dialog.mem_dirs[0]), XtRString, (XtPointer) NULL }
};

static void 
  ClassInitialize _DCL_VOID;
static void 
  Initialize _DCL((new, req), Widget new _AND Widget req);
static void 
  fdUpdateLocation _DCL((dlog), FileDialogPart *dlog );
static void 
  fdScrollTo _DCL((fdw, index),
		  FileDialogWidget fdw _AND int index );
static void 
  fdUpdateFileList _DCL((w, client_data, call_data),
			Widget w _AND XtPointer client_data _AND 
			XtPointer call_data);
static void
  fdChangeDirectory _DCL((w, dlog, fname, check),
			 Widget w _AND FileDialogPart *dlog _AND 
			 char *fname _AND Boolean check);
static void
  fdSelectFile _DCL((w, client_data, call_data),
		    Widget w _AND XtPointer client_data _AND 
		    XtPointer call_data);

#define superclass      (&transientShellClassRec)
FileDialogClassRec fileDialogClassRec = {
  { /* core_class fields */
    /* superclass         */    (WidgetClass) superclass,
    /* class_name         */    "FileDialog",
    /* widget_size        */    sizeof(FileDialogRec),
    /* class_initialize   */    ClassInitialize,
    /* class_part_init    */    NULL,
    /* class_inited       */    FALSE,
    /* initialize         */    Initialize,
    /* initialize_hook    */    NULL,
    /* realize            */    XtInheritRealize,
    /* actions            */    actions,
    /* num_actions        */    XtNumber(actions),
    /* resources          */    resources,
    /* num_resources      */    XtNumber(resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion    */    TRUE,
    /* compress_exposure  */    TRUE,
    /* compress_enterleave*/    TRUE,
    /* visible_interest   */    FALSE,
    /* destroy            */    NULL,
    /* resize             */    XtInheritResize,
    /* expose             */    XtInheritExpose,
    /* set_values         */    NULL,
    /* set_values_hook    */    NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */    NULL,
    /* accept_focus       */    XtInheritAcceptFocus,
    /* version            */    XtVersion,
    /* callback_private   */    NULL,
    /* tm_table           */    translations,
    /* query_geometry     */    XtInheritQueryGeometry,
    /* display_accelerator*/    XtInheritDisplayAccelerator,
    /* extension          */    NULL
  },
  { /* composite_class fields */
    /* geometry_manager   */    XtInheritGeometryManager,
    /* change_managed     */    XtInheritChangeManaged,
    /* insert_child       */    XtInheritInsertChild,
    /* delete_child       */    XtInheritDeleteChild,
    /* extension          */    NULL
  },{
    /* extension          */    NULL
  },{
    /* extension          */    NULL
  },{
    /* extension          */    NULL
  },{
    /* extension          */    NULL
  }
};

WidgetClass fileDialogWidgetClass = (WidgetClass)&fileDialogClassRec;

static FileDialogWidget
findFileDialog _IMPL((w), Widget w )
{
    for ( ; w; w = w->core.parent )
      if ( XtIsSubclass(w,fileDialogWidgetClass) )
	return (FileDialogWidget) w;

    return NULL;
}

static void
actSelect _IMPL((w, ev, params, num_params),
		Widget w _AND XEvent *ev _AND 
		String *params _AND Cardinal *num_params )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;

    dlog->done = DLOG_SELECTED;
}

static void
actCancel _IMPL((w, ev, params, num_params),
		Widget w _AND XEvent *ev _AND 
		String *params _AND Cardinal *num_params )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;
    dlog->done = DLOG_CANCELLED;
}

static void
actToggleLong _IMPL((w, ev, params, num_params),
		    Widget w _AND XEvent *ev _AND 
		    String *params _AND Cardinal *num_params )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;

    dlog->long_listing = !dlog->long_listing;
    dlog->last_mtime = 0;
    fdUpdateFileList((Widget) fdw, (XtPointer) dlog, 0);
}

static void
actToggleFilter _IMPL((w, ev, params, num_params),
		      Widget w _AND XEvent *ev _AND
		      String *params _AND Cardinal *num_params )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;

    dlog->filter = !dlog->filter;
    dlog->last_mtime = 0;
    fdUpdateFileList((Widget) fdw, (XtPointer) dlog, 0);
}

static void
actUpdateFileList _IMPL((w, ev, params, nparams), 
			Widget w _AND XEvent *ev _AND 
			String *params _AND Cardinal *nparams )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;

    fdUpdateFileList((Widget) fdw, (XtPointer)dlog, 0);
}

static void
makeFullname _IMPL((fullname, dname, fname),
		   char *fullname _AND char *dname _AND char *fname )
{
    int len = strlen(dname);

    /* Create the fullname of the new directory. */
    strcpy(fullname,dname);
    if ( dname[len-1] != '/' ) strcat(fullname,"/");
    strcat(fullname,fname);
}

void
XawTextSetString _IMPL((w, name),Widget w _AND String name)
{
    XawTextBlock block;
    String pname = 0;

    XtVaGetValues(w, XtNstring, &pname, NULL);
    block.firstPos = 0;
    block.length = strlen(name);
    block.ptr = name;
    XawTextReplace(w, 0, strlen(pname), &block);
}

static void
actNewDirectory _IMPL((w, ev, params, num_params),
		      Widget w _AND XEvent *ev _AND 
		      String *params _AND Cardinal *num_params )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;
    char *dname = 0, *fname = 0;
    char fullname[MAXPATHLEN];
    struct stat sbuf;

    if ( !dlog->new ) {
	fprintf(stderr,"Cannot mkdir(): Looking for existing files.\n");
	return;
    }

    XtVaGetValues( dlog->dir, XtNstring, &dname, NULL );
    XtVaGetValues( dlog->file_name, XtNstring, &fname, NULL );
    
    /* Only create a subdirectory of a valid directory. */
    if ( stat(dname,&sbuf) < 0 || !S_ISDIR(sbuf.st_mode) ) {
	fprintf(stderr,"Cannot mkdir(): File %s is not directory.\n",dname);
	return;
    }

    makeFullname(fullname,dname,fname);

    /* Don't do it if the file already exists. */
    if ( stat(fullname,&sbuf) == 0 ) {	
	fprintf(stderr,"Cannot mkdir(): File %s already exists.\n",fullname);
	return;
    }

    /* Do it! */
    mkdir(fullname,0777);
    fdChangeDirectory((Widget)fdw, dlog, fullname, True);

    XawTextSetString(dlog->file_name, "");
}

static void
actJumpFile _IMPL((w, ev, params, num_params),
		  Widget w _AND XEvent *ev _AND
		  String *params _AND Cardinal *num_params )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;

    if ( *num_params >= 1 ) {
	XawTextSetString(dlog->dir, params[0]);
	XawTextSetString(dlog->file_name, (*num_params>=2)? params[1]: "");
	fdUpdateLocation(dlog);
	fdUpdateFileList((Widget) fdw, (XtPointer)dlog, 
			 (*num_params>=2)? params[1]: 0 );
    }
}

static void
actNextFile _IMPL((w, ev, params, num_params),
		  Widget w _AND XEvent *ev _AND 
		  String *params _AND Cardinal *num_params )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;
    Widget listWidget;
    Cardinal total;
    int delta = 1;
    XawListReturnStruct *list_r;

    if ( *num_params >= 1 ) delta = atoi(params[0]);
    XtVaGetValues(dlog->file_list, 
		  XtNnumberStrings, &total, XtNlistWidget, &listWidget, 
		  NULL);
    list_r = XawListShowCurrent(listWidget);

    if (list_r->list_index < 0) list_r->list_index = -1;
    list_r->list_index += delta;

    if ( list_r->list_index < 0 ) {
	list_r->list_index = 0;
	XBell(XtDisplay(w),50);
    } else if ( list_r->list_index >= total ) {
	list_r->list_index = total-1;
	XBell(XtDisplay(w),50);
    }

    /* First move w/o callback. */
    fdScrollTo(fdw, list_r->list_index);
    XtFree(list_r);
}

static void
actDeleteFile _IMPL((w, ev, params, num_params), 
		    Widget w _AND XEvent *ev _AND 
		    String *params _AND Cardinal *num_params )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;
    char *dname = 0, *fname = 0;
    char fullname[MAXPATHLEN];
    struct stat sbuf;

    XtVaGetValues( dlog->dir, XtNstring, &dname, NULL );
    XtVaGetValues( dlog->file_name, XtNstring, &fname, NULL );
    if ( !fname || !strcmp(fname,"") ) return;

    makeFullname(fullname,dname,fname);
    if ( stat(fullname,&sbuf) != 0 ) {	
	fprintf(stderr,"Cannot delete: File %s doesn't exist.\n",fname);
	return;
    } else if ( S_ISDIR(sbuf.st_mode) ) {
	fprintf(stderr,"Cannot delete: %s is a directory.\n",fname);
	return;
    }

    if ( ModalPopupConfirm("Delete %s?",fname) ) {
	unlink(fullname);
	fdUpdateFileList((Widget) fdw, (XtPointer)dlog, 0);
    }
}

static void
actSearch _IMPL((w, ev, params, num_params),
		Widget w _AND XEvent *ev _AND 
		String *params _AND Cardinal *num_params )
{
    FileDialogWidget fdw = findFileDialog(w);
    FileDialogPart *dlog = &fdw->dialog;
    KeySym sym = XKeycodeToKeysym(XtDisplay(w),
				  ev->xkey.keycode,
				  ev->xkey.state);
    String *list = 0;
    int i;
    
    XtVaGetValues( dlog->file_list, XtNlist, &list, NULL );
    for ( i = 1; *list; ++list, ++i )
      if ( (*list)[LONG_OFF(dlog)] >= sym ) break;

    if ( *list && (*list)[LONG_OFF(dlog)] == sym )
      XukcScrListSelectItem(dlog->file_list, i, True, True);
}

static void
fdSelected _IMPL((w, client_data, call_data),
		 Widget w _AND XtPointer client_data _AND XtPointer call_data )
{
    FileDialogPart *dlog = &((FileDialogWidget)client_data)->dialog;
    dlog->done = DLOG_SELECTED;
}

static void
fdCancelled _IMPL((w, client_data, call_data),
		  Widget w _AND XtPointer client_data _AND 
		  XtPointer call_data )
{
    FileDialogPart *dlog = &((FileDialogWidget)client_data)->dialog;
    dlog->done = DLOG_CANCELLED;
}

static void
updateList _IMPL((list, length, new, dir),
		 char **list _AND int length _AND char *new _AND int dir)
{
    char **currp, *prev;
    int i, j;
    int len = dir? strrchr(new,'/')-new: 0;
    /* Look for previous occurrence. */
    for ( i = 0, currp = list; *currp && i < length; ++i, ++currp ) {
	if ( (dir && len?
	      !strncmp(*currp,new,len+1) && !strchr(*currp+len+2,'/'):
	      !strcmp(*currp,new)) )
	  break;
    }

    /* If found previous, shuffle to bring to top, otherwise add new. */
    prev = (*currp && i < length)? *currp: strdup(new);
    for ( j = i; j > 0; --j, --currp )
      if ( j < length ) *currp = currp[-1];
      else free(currp[-1]);
    *currp = prev;
}

static char *
strappend _IMPL((a, b, dealloc),
		char *a _AND char *b _AND Boolean dealloc )
{
    if ( !a ) {
	return strdup(b);
    } else {
	int len = strlen(a) + strlen(b) + 1;
	char *new = (char *) malloc(len);
	strcpy(new,a);
	strcat(new,b);
	if ( dealloc ) free(a);
	return new;
    }
}

static char *
strrnchr _IMPL((str, c, n), char *str _AND char c _AND int n )
{
    char *p = str + strlen(str);

    for ( ; p != str; --p )
      if ( *p == c && --n == 0 ) break;

    return p==str? (char*)NULL: p;
}

#define MENU_DELIM "(line)"

static char *
buildFileMenu _IMPL((menu, list, length, as_dir), 
		    char *menu _AND char **list _AND int length _AND 
		    Boolean as_dir )
{
    char menu_entry[MAXPATHLEN];
    char file_name[MAXPATHLEN];
    char indic[MAXPATHLEN];
    char **currp;
    int i;
    Boolean dealloc = False;

    /* 
     * The start of the menu is either right after the last (line), or 
     * at the beginning.
     */
    if ( menu ) {
	char *start = menu, *next;
	for ( next = strstr(menu,MENU_DELIM);
	      next != NULL;
	      next = strstr(start,MENU_DELIM) )
	  start = next + sizeof(MENU_DELIM);

	if ( start ) start[0] = '\0'; else menu = NULL;
    }

    /*
     * Now walk through the entire list of files and build menu entries.
     */
    for ( i = 0, currp = list; *currp && i < length; ++i, ++currp ) {
	char *tail = strrnchr(*currp,'/',as_dir?3:2);
	if ( !tail ) tail = *currp;

	/* Construct args to JumpFile. */
	strcpy(file_name,*currp);
	*strrchr(file_name,'/') = ',';

	if ( as_dir ) {
	    char tbuf[MAXPATHLEN];
	    strcpy(tbuf,tail);
	    *strrchr(tbuf,'/') = '\0';
	    sprintf( indic, "Goto %s%s", tail==*currp? "": "...", tbuf );
	} else {
	    sprintf( indic, "Use %s%s", tail==*currp? "": "...", tail );
	}

	sprintf(menu_entry, 
		" (item \"%s\" (action \"fdJumpFile(%s)\"))\n",
		indic, file_name );
	menu = strappend( menu, menu_entry, dealloc );
	dealloc = True;
    }

    return menu;
}

static void
fdUpdateLocation _IMPL((dlog), FileDialogPart *dlog )
{
    Widget dirMenu  = XtNameToWidget( dlog->dialog, "dirMenu" );
    Widget fileMenu = XtNameToWidget( dlog->dialog, "fileMenu" );
    char *dname = 0, *fname = 0;
    char *dmenu = 0, *fmenu = 0;
    char fullname[MAXPATHLEN];

    XtVaGetValues( dlog->dir, XtNstring, &dname, NULL );
    XtVaGetValues( dlog->file_name, XtNstring, &fname, NULL );
    makeFullname(fullname,dname,fname);

    if ( dirMenu ) {
	int len = strrchr(fullname,'/') - fullname;
	updateList(dlog->mem_dirs, NMEMORY, fullname, len);

	/* Now update the menu. */
	XtVaGetValues( dirMenu, XtNmenu, &dmenu, NULL );
	dmenu = buildFileMenu( dmenu, dlog->mem_dirs, NMEMORY, True );
	XtVaSetValues( dirMenu, XtNmenu, dmenu, NULL );
    }

    if ( fileMenu ) {
	if ( fname && strcmp(fname,"") ) {
	    updateList(dlog->mem_files, NMEMORY, fullname, strlen(fullname));

	    /* Now update the menu. */
	    XtVaGetValues( fileMenu, XtNmenu, &fmenu, NULL );
	    fmenu = buildFileMenu( fmenu, dlog->mem_files, NMEMORY, False );
	    XtVaSetValues( fileMenu, XtNmenu, fmenu, NULL );
	}
    }
}

enum fd_use { FD_DONT_USE=0, FD_USE_DIR=1, FD_USE_FILE=2 };

enum fd_use
fdUseFile _IMPL((dlog, fname, sbuf), 
		FileDialogPart *dlog _AND char *fname _AND struct stat *sbuf )
{
    char *dname = 0;
    char fullname[MAXPATHLEN];
    
    /* Build fullname. */
    XtVaGetValues( dlog->dir, XtNstring, &dname, NULL );
    makeFullname(fullname,dname,fname);

    if ( sbuf && stat(fullname, sbuf) < 0 ) 
      return FD_DONT_USE;
    
    /* If a directory, automatically use it unless it is "." */
    if ( sbuf && S_ISDIR(sbuf->st_mode) && strcmp(fname,".") ) 
      return FD_USE_DIR;

    /* Run through exclusion list if exists. */
    if ( dlog->filter && dlog->exclude ) {
	StringList p = dlog->exclude;
	for ( ; p != NULL; p = p->next ) {
	    if ( regexec(p->regex,fname) ) return FD_DONT_USE;
	}
    }

    /* Run through inclusion list if exists. */
    if ( dlog->filter && dlog->include ) {
	StringList p = dlog->include;
	for ( ; p != NULL; p = p->next ) {
	    if ( regexec(p->regex,fname) ) return FD_USE_FILE;
	}
    }

    /* And finally, call the user-provided function if it exists. */
    if ( dlog->filter && dlog->accept_f ) {
	return (*dlog->accept_f)(fname)? FD_USE_FILE: FD_DONT_USE;
    }

    return FD_USE_FILE;
}

static void
fdScrollTo _IMPL((fdw, index), FileDialogWidget fdw _AND int index )
{
    FileDialogPart *dlog = &fdw->dialog;

    XukcScrListSelectItem(dlog->file_list, index+1, True, True);
}

static void
fdUpdateFileList _IMPL((w, client_data, call_data),
		       Widget w _AND XtPointer client_data _AND 
		       XtPointer call_data )
{
    FileDialogPart *dlog = (FileDialogPart*) client_data;
    String entry_name = (String) call_data;
    struct direct **dlist = 0, **dp;
    char *dname = 0;
    String *list = 0, *plist = 0;
    Cardinal i, j, num, pnum;
    extern int alphasort();
    struct stat sbuf;
    enum fd_use used;
    int list_entry = -1;

    /* Get directory list. */
    XtVaGetValues( dlog->dir, XtNstring, &dname, NULL );

    /* Conditions for bugging out. */
    /* 1) Not a directory. */
    if ( stat(dname,&sbuf) < 0 || !S_ISDIR(sbuf.st_mode) ) 
      return;

    /* Default file to search for is mentioned in the file_name widget. */
    if ( !entry_name )
      XtVaGetValues(dlog->file_name, XtNstring, &entry_name, NULL);

    if ( sbuf.st_ino != dlog->last_ino || sbuf.st_mtime > dlog->last_mtime ) {
	/* Remember previous list... */
	XtVaGetValues(dlog->file_list,
		      XtNlist, &plist, XtNnumberStrings, &pnum,
		      NULL);

	if ( (num=scandir( dname, &dlist, NULL, alphasort )) < 0 )
	  return;

	/* If this is the same one as last time, use current selection. */
	if ( sbuf.st_ino == dlog->last_ino ) {
	    Widget lw;
	    XawListReturnStruct *ret;
	    XtVaGetValues(dlog->file_list, XtNlistWidget, &lw, NULL);
	    ret = XawListShowCurrent(lw);
	    list_entry = ret->list_index;
	} else
	    list_entry = 0;

	/* Remember status. */
	dlog->last_ino = sbuf.st_ino;
	dlog->last_mtime = sbuf.st_mtime;

	/* Format it for use in the fileList widget */
	list = (String *) calloc(num+1, sizeof(String));
	for ( i = 0, j = 0, dp = dlist; i < num; ++i, ++dp ) {
	    if ( (used = fdUseFile(dlog, (*dp)->d_name, &sbuf)) ) {
		char buf[80];
		if ( entry_name && !strcmp(entry_name,(*dp)->d_name) )
		  list_entry = j;
		switch ( used ) {
		  case FD_USE_DIR:
		    if ( dlog->long_listing )
		      sprintf(buf, "----- %s/", (*dp)->d_name);
		    else
		      sprintf(buf, "%s/", (*dp)->d_name);
		    break;
		  case FD_USE_FILE:
		    if ( dlog->long_listing )
		      sprintf(buf, "%4dK %s", 
			      (sbuf.st_size+512)/1024, (*dp)->d_name);
		    else
		      sprintf(buf, "%s", (*dp)->d_name);
		    break;
		}
		list[j++] = strdup(buf);
	    }
	}
	list[j] = 0;
	XukcScrListChange( dlog->file_list, list, j );

	/* Must free previous list here so list != plist! */
	if ( plist && strcmp(plist[0], XtName(dlog->file_list)) ) {
	    for ( i = 0; i < pnum; ++i ) free(plist[i]);
	    free(plist);
	}
    } else if ( entry_name ) {
	/* Find entry in previously existing list. */
	XtVaGetValues(dlog->file_list, 
		      XtNlist, &list, XtNnumberStrings, &j, NULL);
	for ( i = 0; i < j; ++i ) {
	    if ( !strcmp( entry_name, list[i] ) ) {
		list_entry = i;
		break;
	    }
	}
    }

    /* Inform fileList widget of the change. */
    if ( list_entry >= 0 )
      XukcScrListSelectItem(dlog->file_list, list_entry+1, True, True);

    if ( dlist ) free(dlist);
}

static void
fdChangeDirectory _IMPL((w, dlog, fname, check),
			Widget w _AND FileDialogPart *dlog _AND 
			char *fname _AND Boolean check )
{
    char owd[MAXPATHLEN];
    struct stat sbuf;
    Widget file_list = 0;
    extern char *getwd();

    if ( !check || (stat(fname,&sbuf) == 0 && S_ISDIR(sbuf.st_mode)) ) {
	getwd(owd);
	chdir(fname);

	XawTextSetString(dlog->dir, getwd(fname));
	XtVaGetValues( dlog->file_list, XtNlistWidget, &file_list, NULL);
	if ( file_list ) XtVaSetValues( file_list, XtNx, 0, XtNy, 0, NULL);
	fdUpdateFileList( w, (XtPointer)dlog, 0 );
	chdir(owd);
    } else {
	XBell(XtDisplay(w),50);
    }
}

static void
fdSelectFile _IMPL((w, client_data, call_data),
		   Widget w _AND XtPointer client_data _AND 
		   XtPointer call_data )
{
    FileDialogPart *dlog = (FileDialogPart*)client_data;
    XawListReturnStruct *list_elem = (XawListReturnStruct*) call_data;
    char fname[MAXPATHLEN];
    char *dname = 0;
    struct stat sbuf;

    if ( call_data ) {
	String name  = list_elem->string + LONG_OFF(dlog);
	String pname;

	XtVaGetValues( dlog->dir, XtNstring, &dname, NULL );
	strcpy( fname, dname );
	strcat( fname, "/" );
	strcat( fname, name );
	if ( stat(fname,&sbuf) < 0 ) return;

	XawTextSetString(dlog->file_name, name);
    }
}

void
addFileDialogButton _IMPL((w, name, callback),
			  Widget w _AND char *name _AND XtPointer callback )
{
    FileDialogPart *dlog = &((FileDialogWidget)w)->dialog;
    Widget button =
      XtVaCreateManagedWidget(name, commandWidgetClass, dlog->dialog, 
			      XtNfromVert, dlog->file_name,
			      XtNleft, XtChainLeft,
			      XtNright, XtChainLeft,
			      XtNbottom, XtChainBottom,
			      NULL );
    if ( dlog->last_button ) 
      XtVaSetValues( button, XtNfromHoriz, dlog->last_button, NULL );
    XtAddCallback( button, XtNcallback, callback, w );

    dlog->last_button = button;
}

void
addFileDialogExclusion _IMPL((w, pattern), Widget w _AND char *pattern )
{
    FileDialogPart *dlog = &((FileDialogWidget)w)->dialog;
    StringList new = (StringList) malloc( sizeof(struct _StringList) );

    new->string = strdup(pattern);
    new->regex  = regcomp(pattern);
    new->prev   = 0;
    new->next   = dlog->exclude;
    if ( dlog->exclude ) dlog->exclude->prev = new;
    dlog->exclude = new;
}

int
delFileDialogExclusion _IMPL((w, pattern), Widget w _AND char *pattern )
{
    FileDialogPart *dlog = &((FileDialogWidget)w)->dialog;
    StringList p;

    for ( p = dlog->exclude; p != NULL; p = p->next ) {
	if ( !strcmp(pattern,p->string) ) {
	    if ( p->prev ) p->prev->next = p->next;
	    else dlog->exclude = p->next;
	    if ( p->next ) p->next->prev = p->prev;
	    free(p->string);
	    free(p->regex);
	    free(p);
	    return 1;
	}
    }
    return 0;
}

void
addFileDialogInclusion _IMPL((w, pattern), Widget w _AND char *pattern )
{
    FileDialogPart *dlog = &((FileDialogWidget)w)->dialog;
    StringList new = (StringList) malloc( sizeof(struct _StringList) );

    new->string = strdup(pattern);
    new->regex  = regcomp(pattern);
    new->prev   = 0;
    new->next   = dlog->exclude;
    if ( dlog->include ) dlog->include->prev = new;
    dlog->include = new;
}

int
delFileDialogInclusion _IMPL((w, pattern), Widget w _AND char *pattern )
{
    FileDialogPart *dlog = &((FileDialogWidget)w)->dialog;
    StringList p;

    for ( p = dlog->include; p != NULL; p = p->next ) {
	if ( !strcmp(pattern,p->string) ) {
	    if ( p->prev ) p->prev->next = p->next;
	    else dlog->include = p->next;
	    if ( p->next ) p->next->prev = p->prev;
	    free(p->string);
	    free(p->regex);
	    free(p);
	    return 1;
	}
    }
    return 0;
}

void
setFileDialogAcceptFunc _IMPL((w, f), Widget w _AND Boolean (*f) __((String)) )
{
    FileDialogPart *dlog = &((FileDialogWidget)w)->dialog;

    dlog->accept_f = f;
}

static XtTranslations NoTextSearchAndReplace;
static char NoTextSRTransStr[] = 
"Ctrl<Key>R: no-op(RingBell)\n\
 Ctrl<Key>S: no-op(RingBell)\n\
";

static XtTranslations listTrans;
static char listTransStr[] =
"<Btn1Down>,<Btn1Up>: Set() Notify()		\n\
 <Btn1Up>(2):	      Set() Notify() fdSelect() \n\
";

static void
ClassInitialize _IMPL_VOID
{
    XukcRegisterConverters();

    NoTextSearchAndReplace = XtParseTranslationTable(NoTextSRTransStr);
    listTrans = XtParseTranslationTable(listTransStr);
}

static void
augmentFileListTranslations _IMPL((dlog), FileDialogPart *dlog )
{
    Widget file_list;

    /* Install FD-specific translations on the file list. */
    XtVaGetValues(dlog->file_list, XtNlistWidget, &file_list, NULL);
    if ( file_list )
      XtAugmentTranslations(file_list, listTrans);
}

static void
Initialize _IMPL((request, new), Widget request _AND Widget new )
{
    FileDialogWidget fdw = (FileDialogWidget) new;
    FileDialogPart *dlog = &fdw->dialog;
    Widget dialog;
    char dname[MAXPATHLEN];
    int i;
    
    getwd(dname);

    dialog = 
      XtVaCreateManagedWidget("fileDialog", formWidgetClass, new,
			      NULL);

    dlog->dialog = dialog;
    dlog->exclude = 0;
    dlog->include = 0;
    dlog->filter = True;
    dlog->accept_f = 0;
    dlog->last_button = 0;
    dlog->last_ino = 0;
    dlog->last_mtime = 0;
    dlog->placed = False;

    for ( i = 0; i < NMEMORY; ++i ) dlog->mem_files[i] = 0;

    {
	Widget dirMenu =
	  XtVaCreatePopupShell("dirMenu", ezMenuWidgetClass, dialog, NULL);
	Widget dlabel =
	  XtVaCreateManagedWidget("dirLabel", menuButtonWidgetClass, dialog,
				  NULL );
	Widget dir =
	  XtVaCreateManagedWidget("directory", asciiTextWidgetClass, dialog,
				  XtNstring, dname,
				  NULL );

	Widget file_list =
	  XtVaCreateManagedWidget("fileList", ukcScrListWidgetClass, dialog, 
				  NULL);
	  
	Widget fileMenu =
	  XtVaCreatePopupShell("fileMenu", ezMenuWidgetClass, dialog, NULL);
	Widget flabel =
	  XtVaCreateManagedWidget("fileLabel", menuButtonWidgetClass, dialog,
				  NULL );
	Widget file_name =
	  XtVaCreateManagedWidget("fileName", asciiTextWidgetClass, dialog,
				  NULL );

	dlog->dir = dir;
	dlog->file_name = file_name;
	dlog->file_list = file_list;

	/* Don't search/replace in directory and file strings. */
	XtOverrideTranslations(dir, NoTextSearchAndReplace);
	XtOverrideTranslations(file_name, NoTextSearchAndReplace);

	XtAddCallback(file_list, XtNcallback, fdSelectFile, dlog );

	/* Must add callback to asciiSrc object, so must access element. */
	{
	    Widget asciiSrc;
	    XFontStruct *font;
	    Dimension width;

	    XtVaGetValues(dir, XtNtextSource, &asciiSrc, 
			  XtNfont, &font, NULL );
	    width = XTextWidth(font,dname,strlen(dname));
	    XtVaSetValues(dir,XtNwidth,width+30,NULL);

	    XtAddCallback(asciiSrc, XtNcallback, fdUpdateFileList, dlog );
	}

	/* Translate default search directories to dirMenu. */
	if ( dlog->mem_dirs[0] ) {
	    char *dbuf, *p, *n;
	    dbuf = strdup(dlog->mem_dirs[0]);
	    for ( i = 0, p = dbuf, n = p? strchr(p,':'):0;
		  p && i < NMEMORY;
		  n? (p = n+1, n = strchr(p,':')): (p=0) ) {
		if ( n ) *n = '\0';
		/* Special case.  Empty entry means sub in BITMAPPATH. */
		if ( *p == 0 ) {
		} else {
		    char *space = malloc( strlen(p) + 2 );
		    strcpy( space, p );
		    strcat( space, "/" );
		    dlog->mem_dirs[i++] = space;
		}
	    }
	    for ( ; i < NMEMORY; ++i ) dlog->mem_dirs[i] = 0;
	    free(dbuf);
	}
    }

    /* Create select and cancel buttons. */
    addFileDialogButton(new, "selectButton", (XtPointer)fdSelected);
    addFileDialogButton(new, "cancelButton", (XtPointer)fdCancelled);

    /* Exclude invisible and temporary files. */
    addFileDialogExclusion(new,"^\\.");
    addFileDialogExclusion(new,"^#");
    addFileDialogExclusion(new,"~$");

    fdUpdateLocation(dlog);
}

static Widget
findShell _IMPL((w), Widget w )
{
    for ( ; w; w = w->core.parent )
      if ( XtIsShell(w) ) return w;

    return NULL;
}

Widget
makeFileDialog _IMPL((name, parent), char *name _AND Widget parent )
{
    Widget shell = findShell(parent);
    Visual *vis;
    Colormap cmap;

    XtVaGetValues( shell, XtNvisual, &vis, XtNcolormap, &cmap, NULL );
    return 
      XtVaCreatePopupShell(name, fileDialogWidgetClass, parent,
			   XtNvisual, vis,  XtNcolormap, cmap,
			   NULL);
}

char *
useFileDialog _IMPL((w, new), Widget w _AND Boolean new )
{
    FileDialogPart *dlog = &((FileDialogWidget)w)->dialog;
    XtAppContext app_context = XtWidgetToApplicationContext(w);
    Position x, y;
    Dimension width, height;

    /* Ensure that file list up to date. */
    fdUpdateFileList(dlog->file_list, (XtPointer)dlog, 0);
    augmentFileListTranslations(dlog);
    XtRealizeWidget(w);

    /* 
     * Either initially place the window under the mouse, or move the 
     * mouse over the window.
     */
    if ( !dlog->placed ) {
	Window r, c;
	int rx, ry, cx, cy;
	unsigned int m;
	Dimension p_width, p_height, bw;

	/* Resize the file list to the dialog width. */
	XtVaGetValues(w, XtNwidth, &width, XtNheight, &height, NULL); 
	XtVaGetValues(dlog->file_list, XtNx, &x, XtNwidth, &p_width,
		      XtNheight, &p_height, XtNborderWidth, &bw, NULL);
	if ( width - 2*bw > p_width ) {
	    XtVaSetValues(dlog->file_list, XtNresizable, True, NULL);
	    XtVaSetValues(dlog->file_list, XtNwidth, width-2*(x+bw), NULL);
	    XtVaSetValues(dlog->file_list, XtNresizable, False, NULL);
	}

	/* Centre the widget over the pointer. */
	XQueryPointer(XtDisplay(w), XtWindow(w), &r,&c, &rx,&ry,&cx,&cy, &m);
	x = rx-width/2;  if ( x < 0 ) x = 0;
	y = ry-height/2; if ( y < 0 ) y = 0;
	XtVaSetValues(w, XtNx, x, XtNy, y, NULL);
	dlog->placed = True;
    } else {
	XtVaGetValues(w, XtNwidth, &width, XtNheight, &height, NULL); 
	XWarpPointer( XtDisplay(w), NULL, XtWindow(w), 0,0,0,0, 
		      width/2, height/2 );
    }

    /* Ensure that actions are propagated down. */
    XtInstallAcceleratorsBelow(w, w);
    XtInstallAcceleratorsBelow(dlog->file_list, dlog->file_list);
    XtInstallAllAccelerators(dlog->file_name, dlog->file_name);

    XtPopup( w, XtGrabNone );

    dlog->done = DLOG_NOTDONE;
    dlog->new = new;
    while ( dlog->done == DLOG_NOTDONE ) {
	XEvent event;
	XtAppNextEvent(app_context,&event);
        XtDispatchEvent(&event);
	if ( dlog->done == DLOG_SELECTED ) {
	    char fullname[MAXPATHLEN];
	    char *dname = 0, *fname = 0;
	    struct stat sbuf;

	    XtVaGetValues( dlog->dir, XtNstring, &dname, NULL );
	    XtVaGetValues( dlog->file_name, XtNstring, &fname, NULL );
	    if ( stat(dname,&sbuf) < 0 || !S_ISDIR(sbuf.st_mode)
		 || *fname == '\0' ) {
		/* 
		 * An invalid directory entry or file doesn't allow one to
		 * select the file.   Beep and try again.
		 */
		dlog->done = DLOG_NOTDONE;
		XBell(XtDisplay(w),50);
	    } else {
		makeFullname(fullname,dname,fname);
		if ( !new && stat(fullname,&sbuf) < 0 ) {
		    dlog->done = DLOG_NOTDONE;
		    XBell(XtDisplay(w),50);
		} else if ( S_ISDIR(sbuf.st_mode) ) {
		    dlog->done = DLOG_NOTDONE;
		    fdChangeDirectory(w,dlog,fullname,False);
		} else {	
		    fdUpdateLocation(dlog);
		    XtPopdown(w);
		    XSync(XtDisplay(w),False);
		    return strdup(fullname);
		}
	    }
	}
    }

    XtPopdown(w);
    XSync(XtDisplay(w),False);
    return 0;
}
