/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

   $Id: Image.h,v 0.9.1.2 1991/11/17 19:28:21 leei Exp $

******************************************************************************/

/* Derived from */
/* $XConsortium: Template.h,v 1.4 89/07/21 01:41:49 kit Exp $ */
/* Copyright	Massachusetts Institute of Technology	1987, 1988 */

#ifndef _Image_h
#define _Image_h

/****************************************************************
 *
 * Image widget
 *
 ****************************************************************/

/* Resources:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 background	     Background		Pixel		XtDefaultBackground
 border		     BorderColor	Pixel		XtDefaultForeground
 borderWidth	     BorderWidth	Dimension	1
 description	     Description	Boolean		False
 destroyCallback     Callback		Pointer		NULL
 displayN	     DisplayN		Int		1
 font		     Font		XFontStruct	XtDefaultFont
 foreground	     Foreground		Pixel		XtDefaultForeground
 grid		     Grid		Boolean		False
 gridFont	     Font		XFontStruct	XtDefaultFont
 height		     Height		Dimension	0
 imageWidth	     ImageWidth		Dimension	0
 imageHeight	     ImageHeight	Dimension	0
 mappedWhenManaged   MappedWhenManaged	Boolean		True
 redrawsDelayed	     RedrawsDelayed	Boolean		False
 sensitive	     Sensitive		Boolean		True
 verbose	     Verbose		Boolean		False
 width		     Width		Dimension	0
 x		     Position		Position	0
 y		     Position		Position	0
*/

/* define any special resource names here that are not in <X11/StringDefs.h> */
#define XtNdescription "description"
#define XtNdisplayN "displayN"
#define XtNgrid "grid"
#define XtNgridFont "gridFont"
#define XtNimageWidth "imageWidth"
#define XtNimageHeight "imageHeight"
#define XtNmagX "magX"
#define XtNmagY "magY"
#define XtNredrawsDelayed "redrawsDelayed"
#define XtNstackLength "stackLength"
#define XtNstack "stack"
#define XtNverbose "verbose"

#define XtCDescription "Description"
#define XtCDisplayN "DisplayN"
#define XtCGrid "Grid"
#define XtCImageWidth "ImageWidth"
#define XtCImageHeight "ImageHeight"
#define XtCMag "Magnification"
#define XtCRedrawsDelayed "RedrawsDelayed"
#define XtCStackLength "StackLength"
#define XtCStack "Stack"
#define XtCVerbose "Verbose"

#define XtRDisplayParams "DisplayParams"

/* declare specific ImageWidget class and instance datatypes */

typedef struct _ImageWidgetClassRec*	ImageWidgetClass;
typedef struct _ImageWidgetRec*		ImageWidget;

/* declare the class constant */

extern WidgetClass imageWidgetClass;

/* Now declare functions for manipulating these. */

#include <vim/proto.h>

#include <vim/GenImage.h>
#include <vim/ImViewport.h>

ImageWidget findImageWidget __((Widget w, Boolean error));
ImageViewportWidget findImageViewport __((Widget w, Boolean error));

void
  ImDelayRedraws __((ImageWidget w));
void
  ImRedraw __((ImageWidget w, Boolean force));

Boolean
  ImPushImageStack __((ImageWidget w, GenericImage im));
void
  ImPopImageStack __((ImageWidget w, Boolean redraw, Boolean destroy));
void
  ImRotImageStack __((ImageWidget w, int n, Boolean redraw));

Boolean
  ImSelectBox __((ImageWidget w, int *x1, int *y1, int *x2, int *y2, 
		  XEvent *sev));
Boolean
  ImSelectLine __((ImageWidget w, int *x1, int *y1, int *x2, int *y2, 
		   XEvent *sev));

void
  ImSetMenu __((Widget ww, char *menu_name, char *menu_descr, char *pat, 
		int occurrence, Boolean trunc));

Widget
  ImCreateStackEditor __((Widget imw));

#endif /* _Image_h */
