.\" $Id: vimage.man,v 0.9.1.1 1991/12/12 21:07:06 leei Exp $
.TH vimage 1
.SH NAME
\fBvimage\fP \- displays and manipulates images on X11 displays
.SH SYNTAX
\fBvimage\fP \fI[options] [filename [filename...]]\fP
.SH DESCRIPTION
\fIvimage\fP is an X11 program which displays and manipulates images for X11.
It can access a blinding variety of formats (anything understood by
the PBMplus package), can zoom, scroll, dither, etc.  It is easily
user-extensible and customizable.  In short, it's just what you're
looking for!
.SH OVERVIEW
\fIvimage\fP is essentially just a program-level wrapper for a general
image display widget.  This widget (and it's viewport) supports
sophisticated, interactive display of images.   Images may be scaled,
panned, overlayed and gamma-corrected both under interactive and
program control.
.PP
The Image widget itself is the core of the implementation.  It can be
used in other programs and can be program-controlled.  It consists of
a manipulation system and an interface to a general.
.PP
The Image widget supports the display of a variety of image types.
In loading images, it examines the first few bytes of the file and
attempts to classify it based on those.  If that is insufficient it
then checks against standard extensions.  For an image recognized
by the \fIPBMplus\fP utilities, a pipe is created to to convert it to one of
the internally recognized types.  This process is potentially
recursive, so (for example) a btoa'ed, compressed, Sun raster
file will be unbtoa'ed, uncompressed and then rasttopnm'ed before it
is finally read as a pbm file!  This should operate automatically.
.PP
Images are internally represented as instances of an object-oriented
Xt-like class heirarchy.  All such images are subclasses of the
GenericImage class, which implements most of the interface with the
Image widget.  Only three non-generic image types come with the \fIvimage\fP
package: BitImage (b/w bitmaps), ByteImage (color-mapped 8bit images),
and RGBImage (true-color 24bit images).
.PP
The final twist is the definition of another file meta-type \fIgim\fP,
or `generic image'.  This type consists of a text header followed by
arbitrary data.  The first line of the header names the file type,
which is then used to dynamically load the image reading and
manipulation functions from a user-specified location.  Combining this
facility with the Image Class heirarchy described above, a
user may locally and transparently define image types specific to an
application (e.g. flow fields, edge maps, etc.).
.SH USAGE
Once \fIvimage\fP is started, it is relatively simple (and hopefully
intuitive) to use.  The window layout consists of one or more image
viewport(s).  Each viewport supports a \fIstack\fP of images.  In
normal viewing situations this stack will be one image deep (say
`monalisa.gif'), but when using transparent image types (e.g. a flow
field), this may have the effect of overlaying the display of one
image over another!
.PP
Whenever the cursor is moved into the top-left corner of
a viewport, a menu-bar appears: \fBImage File Stack User\fP.  These
are pull-down menus.  The \fBImage\fP menu contains the basic image
manipulation functions (zoom, redraw, etc.).  The \fBFile\fP menu
contains file-system interface actions.  The most common of these in
normal use is ``Replace Image (^N)'', which deletes the currently
displayed image and requests a replacement using the File Dialog box.
The \fBStack\fP menu consists of functions for manipulating the stack
of images in the current viewport.  The \fBUser\fP menu is a
repository for user extensions.
.PP
In each of these menus (including submenus), commands with keyboard
shortcuts show those shortcuts on the right-hand side of the menu
(e.g. Redraw can be invoked with ^L).  In general, any additions to
these menus should attempt to follow the same convention.
.PP
.SH COMMAND LINE OPTIONS
All of the standard Xt options are supported (see X(1)), as well as
the following:
.IP "\fB\-vpconf\fP \fIn\fPx\fIm\fP" 5
Create \fIn\fPx\fIm\fP independent image viewports in the main window.
The window is configured with \fIn\fP across and \fIm\fP down.
.IP "\fB\-dual\fP, \fB\-quad\fP" 5
Specify a viewport configuration of \fI2x1\fP or \fI2x2\fP respectively.
.IP "\fB\-vps\fP \fIn\fP" 5
Specify a (nearly) square viewport configuration with \fIn\fP viewports.
.IP "\fB\-visual\fP \fIvisual-type\fP, \fB\-mono\fP, \fB\-grey\fP, \fB\-true\fP" 5
Specifies the preferred visual type for the window.
.IP "\fB\-debug\fP \fI{imMdcl}\fP" 5
Sets debugging flags for verbose operation.  The single-letter
arguments stand for different phases of operation and may be combined
arbitrarily: \fIi\fP \(em image widget operation, \fIm\fP \(em memory
allocation, \fIM\fP \(em image method invocation, \fId\fP \(em display
of images, \fIc\fP \(em color table management, and \fIl\fP \(em
dynamic loading.  So \fI-debug dM\fP requests both display and method
debugging.
.IP "\fB\-[no]mtrace\fP, \fB\-[no]mcheck\fP" 5
Enables/disables GNU memory allocation tracing/checking.
.PP
.SH FILE FORMATS

.PP
.SH SHORTCUTS
Commands and their keyboard equivalents (`_' represents the Shift key!):
.in +0.5i
.ta 3in
.nf
\(em \fBImage Menu\fP
.br
Quit:	`q', `^C'
.br
Edit Display Params	`^D'
.br
Edit Image Params	`^E'
.br
Toggle Grid	`G'
.br
Redraw	`^L'
.br
\(em \fBImage Zoom submenu\fP
.br
Reset Zoom	`^Z'
.br
Zoom In	`Z'
.br
Zoom Out	`_Z'
.br
Fit To Window	`^F'
.br
Zoom Region	`^_Z'
.br
Resize Window	`^R'
.br
\(em \fBFile Menu\fP
.br
Replace Image	`^N'
.br
Push Image	`^P'
.br
\(em \fBStack Menu\fP
.br
Edit Stack	`^S'
.br
Replace Image	`^N'
.br
Push Image	`^P'
.br
Pop Image	`^_P'
.br
Rotate Forward	`R'
.br
Rotate Backward	`_R'
.br
.ta
.in -0.5i
.fi
.PP
.SH LIMITATIONS
The system has only been extensively tested on Sun4's under SunOS 4.?,
and a VAXstation under Ultrix 4.0.  It operates well on both 1-bit and
8-bit displays (should or may work on a 24-bit truecolor display).
.PP
Because of limitations in DLD (the GNU dynamic loader), this program
cannot be linked dynamically under SunOS.
.SH KNOWN BUGS
The Viewport Menu Bar pops up whenever the viewport is remapped.
.PP
Has some problems with dynamic loading of library files under Ultrix.
.PP
Reloading an image file in the Image Stack dialog \fBwill\fP cause a crash.
.SH BUG REPORTS
Should be sent to the author, listed below.  Please include the system
type and OS version, along with a complete description of the crash,
bug or whatever.
.SH AUTHOR
Lee Iverson  -  leei@McRCIM.McGill.EDU
.PP
The display code was freely adapted from Jim Frost's xloadimage code.
Nice stuff, and amazingly adaptable.
.PP
GIF reading code based on Jim Frost's adaptation of Kirk Johnson's
<tuna@athena.mit.edu> code.

